@@ -60,10 +60,10 @@
 		      * require consumer to be updated (as long as this number
 		      * is zero, the API is not considered stable and can
 		      * change without a bump of the major version) */
-#define MINVERSION 6 /* API compatible, ABI may change, functional
+#define MINVERSION 7 /* API compatible, ABI may change, functional
 		      * enhancements only, consumer can be left unchanged if
 		      * enhancements are not considered */
-#define PATCHLEVEL 5 /* API / ABI compatible, no functional changes, no
+#define PATCHLEVEL 0 /* API / ABI compatible, no functional changes, no
 		      * enhancements, bug fixes only */
 
 /* remove once in if_alg.h */
@@ -105,7 +105,7 @@ static ssize_t _kcapi_common_send_meta(struct kcapi_handle *handle,
 	/* IV data */
 	struct af_alg_iv *alg_iv = NULL;
 	size_t iv_msg_size = handle->cipher.iv ?
-			  CMSG_SPACE(sizeof(*alg_iv) + handle->cipher.ivlen) :
+			  CMSG_SPACE(sizeof(*alg_iv) + handle->info.ivsize) :
 			  0;
 
 	/* AEAD data */
@@ -144,8 +144,8 @@ static ssize_t _kcapi_common_send_meta(struct kcapi_handle *handle,
 		header->cmsg_type = ALG_SET_IV;
 		header->cmsg_len = iv_msg_size;
 		alg_iv = (void*)CMSG_DATA(header);
-		alg_iv->ivlen = handle->cipher.ivlen;
-		memcpy(alg_iv->iv, handle->cipher.iv, handle->cipher.ivlen);
+		alg_iv->ivlen = handle->info.ivsize;
+		memcpy(alg_iv->iv, handle->cipher.iv, handle->info.ivsize);
 	}
 
 	/* set AEAD information */
@@ -256,22 +256,6 @@ static inline int _kcapi_common_setkey(struct kcapi_handle *handle,
 	return (ret >= 0) ? ret : -errno;
 }
 
-static inline int _kcapi_common_setiv(struct kcapi_handle *handle,
-				      const unsigned char *iv, size_t ivlen)
-{
-	int cipher_ivlen = handle->info.ivsize;
-
-	if (cipher_ivlen < 0)
-		return cipher_ivlen;
-	if (!iv || ivlen != (size_t)cipher_ivlen)
-		return -EINVAL;
-
-	handle->cipher.iv = iv;
-	handle->cipher.ivlen = ivlen;
-
-	return 0;
-}
-
 static int __kcapi_common_getinfo(struct kcapi_handle *handle,
 				  const char *ciphername,
 				  int drivername)
@@ -689,37 +673,17 @@ int kcapi_cipher_setkey(struct kcapi_handle *handle,
 	return _kcapi_common_setkey(handle, key, keylen);
 }
 
-/**
- * kcapi_cipher_setiv() -set IV for the cipher operation
- * @handle: cipher handle - input
- * @iv: buffer holding the IV (may be NULL if IV is not needed) - input
- * @ivlen: length of iv (should be zero if iv is NULL) - input
- *
- * This function requires IV to be exactly IV size. The function verifies
- * the IV size to avoid unnecessary kernel round trips.
- *
- * Note - the caller must keep the iv buffer for every kcapi_cipher_encrypt() or
- * kcapi_cipher_decrypt() operation, or until the last invocation of
- * kcapi_cipher_stream_init_enc() or kcapi_cipher_stream_init_dec() is
- * performed.
- *
- * Return: 0 upon success;
- *	   < 0 in case of an error
- *
- */
-int kcapi_cipher_setiv(struct kcapi_handle *handle,
-		       const unsigned char *iv, size_t ivlen)
-{
-	return _kcapi_common_setiv(handle, iv, ivlen);
-}
-
 /**
  * kcapi_cipher_encrypt() - encrypt data (one shot)
  * @handle: cipher handle - input
  * @in: plaintext data buffer - input
  * @inlen: length of in buffer - input
+ * @iv: IV to be used for cipher operation - input
  * @out: ciphertext data buffer - output
  * @outlen: length of out buffer - input
+ * @access: kernel access type (KCAPI_ACCESS_HEURISTIC - use internal heuristic
+ *	    for  fastest kernel access; KCAPI_ACCESS_VMSPLICE - use vmsplice
+ *	    access; KCAPI_ACCESS_SENDMSG - sendmsg access)
  *
  * It is perfectly legal to use the same buffer as the plaintext and
  * ciphertext pointers. That would mean that after the encryption operation,
@@ -729,12 +693,15 @@ int kcapi_cipher_setiv(struct kcapi_handle *handle,
  * posix_memalign(PAGE_SIZE), If it is not aligned at the page boundary,
  * the vmsplice call may not send all data to the kernel.
  *
+ * The IV buffer must be exactly kcapi_cipher_ivsize() bytes in size.
+ *
  * Return: number of bytes encrypted upon success;
  *	   < 0 in case of error with errno set
  */
 ssize_t kcapi_cipher_encrypt(struct kcapi_handle *handle,
 			     const unsigned char *in, size_t inlen,
-			     unsigned char *out, size_t outlen)
+			     const unsigned char *iv,
+			     unsigned char *out, size_t outlen, int access)
 {
 	struct iovec iov;
 	ssize_t ret = 0;
@@ -755,16 +722,20 @@ ssize_t kcapi_cipher_encrypt(struct kcapi_handle *handle,
 		return -EINVAL;
 	}
 
-	iov.iov_base = (void*)(uintptr_t)in;
+	handle->cipher.iv = iv;
 
+	iov.iov_base = (void*)(uintptr_t)in;
 	/*
 	 * Using two syscalls with memcpy is faster than four syscalls
 	 * without memcpy below the given threshold.
 	 */
-	if (inlen <= (1<<13)) {
+	if ((access == KCAPI_ACCESS_HEURISTIC && inlen <= (1<<13)) ||
+	    access == KCAPI_ACCESS_SENDMSG) {
 		iov.iov_len = inlen;
 		ret = _kcapi_common_send_meta(handle, &iov, 1, ALG_OP_ENCRYPT,
 					      0);
+		if (0 > ret)
+			return ret;
 		iov.iov_base = (void*)(uintptr_t)out;
 		iov.iov_len = outlen;
 		return _kcapi_common_recv_data(handle, &iov, 1);
@@ -773,7 +744,6 @@ ssize_t kcapi_cipher_encrypt(struct kcapi_handle *handle,
 	ret = _kcapi_common_send_meta(handle, NULL, 0, ALG_OP_ENCRYPT, 0);
 	if (0 > ret)
 		return ret;
-
 	while (inlen) {
 		size_t datalen = (inlen > MAXPIPELEN) ? MAXPIPELEN : inlen;
 
@@ -795,8 +765,12 @@ ssize_t kcapi_cipher_encrypt(struct kcapi_handle *handle,
  * @handle: cipher handle - input
  * @in: ciphertext data buffer - input
  * @inlen: length of in buffer - input
+ * @iv: IV to be used for cipher operation - input
  * @out: plaintext data buffer - output
  * @outlen: length of out buffer - input
+ * @access: kernel access type (KCAPI_ACCESS_HEURISTIC - use internal heuristic
+ *	    for fastest kernel access; KCAPI_ACCESS_VMSPLICE - use vmsplice
+ *	    access; KCAPI_ACCESS_SENDMSG - sendmsg access)
  *
  * It is perfectly legal to use the same buffer as the plaintext and
  * ciphertext pointers. That would mean that after the encryption operation,
@@ -806,12 +780,15 @@ ssize_t kcapi_cipher_encrypt(struct kcapi_handle *handle,
  * posix_memalign(PAGE_SIZE), If it is not aligned at the page boundary,
  * the vmsplice call may not send all data to the kernel.
  *
+ * The IV buffer must be exactly kcapi_cipher_ivsize() bytes in size.
+ *
  * Return: number of bytes decrypted upon success;
  *	   < 0 in case of error with errno set
  */
 ssize_t kcapi_cipher_decrypt(struct kcapi_handle *handle,
 			     const unsigned char *in, size_t inlen,
-			     unsigned char *out, size_t outlen)
+			     const unsigned char *iv,
+			     unsigned char *out, size_t outlen, int access)
 {
 	struct iovec iov;
 	ssize_t ret = 0;
@@ -838,13 +815,16 @@ ssize_t kcapi_cipher_decrypt(struct kcapi_handle *handle,
 		return -EINVAL;
 	}
 
+	handle->cipher.iv = iv;
+
 	iov.iov_base = (void*)(uintptr_t)in;
 
 	/*
 	 * Using two syscalls with memcpy is faster than four syscalls
 	 * without memcpy below the given threshold.
 	 */
-	if (inlen <= (1<<13)) {
+	if ((access == KCAPI_ACCESS_HEURISTIC && inlen <= (1<<13)) ||
+	    access == KCAPI_ACCESS_SENDMSG) {
 		iov.iov_len = inlen;
 		ret = _kcapi_common_send_meta(handle, &iov, 1, ALG_OP_DECRYPT,
 					      0);
@@ -876,6 +856,7 @@ ssize_t kcapi_cipher_decrypt(struct kcapi_handle *handle,
 /**
  * kcapi_cipher_stream_init_enc() - start an encryption operation (stream)
  * @handle: cipher handle - input
+ * @iv: IV to be used for cipher operation - input
  * @iov: scatter/gather list with data to be encrypted. This is the pointer to
  *	 the first @iov entry if an array of @iov entries is supplied. See
  *	 sendmsg(2) for details on how @iov is to be used. This pointer may be
@@ -895,12 +876,16 @@ ssize_t kcapi_cipher_decrypt(struct kcapi_handle *handle,
  * kcapi_cipher_stream_op() can be mixed, even by multiple threads of an
  * application.
  *
+ * The IV buffer must be exactly kcapi_cipher_ivsize() bytes in size.
+ *
  * Return: number of bytes sent to the kernel upon success;
  *	   < 0 in case of error with errno set
  */
 ssize_t kcapi_cipher_stream_init_enc(struct kcapi_handle *handle,
+				     const unsigned char *iv,
 				     struct iovec *iov, size_t iovlen)
 {
+	handle->cipher.iv = iv;
 	return _kcapi_common_send_meta(handle, iov, iovlen, ALG_OP_ENCRYPT,
 				       MSG_MORE);
 }
@@ -908,6 +893,7 @@ ssize_t kcapi_cipher_stream_init_enc(struct kcapi_handle *handle,
 /**
  * kcapi_cipher_stream_init_dec() - start a decryption operation (stream)
  * @handle: cipher handle - input
+ * @iv: IV to be used for cipher operation - input
  * @iov: scatter/gather list with data to be decrypted. This is the pointer to
  *	 the first @iov entry if an array of @iov entries is supplied. See
  *	 sendmsg(2) for details on how @iov is to be used. This pointer may be
@@ -927,12 +913,16 @@ ssize_t kcapi_cipher_stream_init_enc(struct kcapi_handle *handle,
  * kcapi_cipher_stream_op() can be mixed, even by multiple threads of an
  * application.
  *
+ * The IV buffer must be exactly kcapi_cipher_ivsize() bytes in size.
+ *
  * Return: number of bytes sent to the kernel upon success;
  *	   < 0 in case of error with errno set
  */
 ssize_t kcapi_cipher_stream_init_dec(struct kcapi_handle *handle,
+				     const unsigned char *iv,
 				     struct iovec *iov, size_t iovlen)
 {
+	handle->cipher.iv = iv;
 	return _kcapi_common_send_meta(handle, iov, iovlen, ALG_OP_DECRYPT,
 				       MSG_MORE);
 }
@@ -1079,28 +1069,6 @@ int kcapi_aead_setkey(struct kcapi_handle *handle,
 	return _kcapi_common_setkey(handle, key, keylen);
 }
 
-/**
- * kcapi_aead_setiv() - set IV for the AEAD operation
- * @handle: cipher handle - input
- * @iv: buffer holding the IV (may be NULL if IV is not needed) - input
- * @ivlen: length of iv (should be zero if iv is NULL) - input
- *
- * Return: 0 upon success; < 0 in case of an error
- *
- * This function requires IV to be exactly IV size. The function verifies
- * the IV size to avoid unnecessary kernel round trips.
- *
- * Note - the caller must keep the iv buffer for every kcapi_aead_encrypt() or
- * kcapi_aead_decrypt() operation, or until the last invocation of
- * kcapi_aead_stream_init_enc() or kcapi_aead_stream_init_dec() is performed.
- * This function only retains a pointer to the user-provided IV buffer.
- */
-int kcapi_aead_setiv(struct kcapi_handle *handle,
-		     const unsigned char *iv, size_t ivlen)
-{
-	return _kcapi_common_setiv(handle, iv, ivlen);
-}
-
 /**
  * kcapi_aead_settaglen() - Set authentication tag size
  * @handle: cipher handle - input
@@ -1144,9 +1112,13 @@ void kcapi_aead_setassoclen(struct kcapi_handle *handle, size_t assoclen)
  * @handle: cipher handle - input
  * @in: plaintext data buffer - input
  * @inlen: length of plaintext buffer - input
+ * @iv: IV to be used for cipher operation - input
  * @assoc: associated data of size set with kcapi_aead_setassoclen() - input
  * @out: data buffer holding cipher text and authentication tag - output
  * @outlen: length of out buffer - input
+ * @access: kernel access type (KCAPI_ACCESS_HEURISTIC - use internal heuristic
+ *	    for fastest kernel access; KCAPI_ACCESS_VMSPLICE - use vmsplice
+ *	    access; KCAPI_ACCESS_SENDMSG - sendmsg access)
  *
  * The AEAD cipher operation requires the furnishing of the associated
  * authentication data. In case such data is not required, it can be set to
@@ -1160,6 +1132,8 @@ void kcapi_aead_setassoclen(struct kcapi_handle *handle, size_t assoclen)
  * posix_memalign(PAGE_SIZE), If it is not aligned at the page boundary,
  * the vmsplice call may not send all data to the kernel.
  *
+ * The IV buffer must be exactly kcapi_cipher_ivsize() bytes in size.
+ *
  * After invoking this function the caller should use
  * kcapi_aead_getdata() to obtain the resulting ciphertext and authentication
  * tag references.
@@ -1169,8 +1143,9 @@ void kcapi_aead_setassoclen(struct kcapi_handle *handle, size_t assoclen)
  */
 ssize_t kcapi_aead_encrypt(struct kcapi_handle *handle,
 			   const unsigned char *in, size_t inlen,
+			   const unsigned char *iv,
 			   const unsigned char *assoc, unsigned char *out,
-			   size_t outlen)
+			   size_t outlen, int access)
 {
 	struct iovec iov[2];
 	ssize_t ret = 0;
@@ -1186,9 +1161,10 @@ ssize_t kcapi_aead_encrypt(struct kcapi_handle *handle,
 		return -EINVAL;
 	}
 
-#if 0
-	if (inlen + handle->aead.assoclen >= (1<<17)) {
-#endif
+	handle->cipher.iv = iv;
+
+	if (access == KCAPI_ACCESS_HEURISTIC ||
+	    access == KCAPI_ACCESS_SENDMSG) {
 		if (assoc && handle->aead.assoclen) {
 			iov[0].iov_base = (void*)(uintptr_t)assoc;
 			iov[0].iov_len = handle->aead.assoclen;
@@ -1204,14 +1180,14 @@ ssize_t kcapi_aead_encrypt(struct kcapi_handle *handle,
 		}
 		if (0 > ret)
 			return ret;
-		iov[0].iov_base = (void*)(uintptr_t)out;
-		iov[0].iov_len = outlen;
-		return _kcapi_common_recv_data(handle, &iov[0], 1);
-
-#if 0
+		ret = _kcapi_common_read_data(handle, out, outlen);
+		if (ret < 0)
+			return ret;
+		if ((ret < (ssize_t)handle->aead.taglen))
+			return -E2BIG;
+		return ret;
 	}
 
-	/* using separate pages slows down splice below the speed of sendmsg */
 	ret = _kcapi_common_send_meta(handle, NULL, 0, ALG_OP_ENCRYPT,
 				      MSG_MORE);
 	if (0 > ret)
@@ -1237,16 +1213,13 @@ ssize_t kcapi_aead_encrypt(struct kcapi_handle *handle,
 	if (0 > ret)
 		return ret;
 
-	iov[0].iov_base = (void*)(uintptr_t)out;
-	iov[0].iov_len = outlen;
-	ret = _kcapi_common_recv_data(handle, &iov[0], 1);
+	ret = _kcapi_common_read_data(handle, out, outlen);
 	if (ret < 0)
 		return ret;
 	if ((ret < (ssize_t)handle->aead.taglen))
 		return -E2BIG;
 
 	return ret;
-#endif
 }
 
 /**
@@ -1294,10 +1267,14 @@ void kcapi_aead_getdata(struct kcapi_handle *handle,
  * @handle: cipher handle - input
  * @in: ciphertext data buffer - input
  * @inlen: length of in buffer - input
+ * @iv: IV to be used for cipher operation - input
  * @assoc: associated data of size set with kcapi_aead_setassoclen() - input
  * @tag: authentication tag data of size set with kcapi_aead_settaglen() - input
  * @out: plaintext data buffer - output
  * @outlen: length of out buffer - input
+ * @access: kernel access type (KCAPI_ACCESS_HEURISTIC - use internal heuristic
+ *	    for fastest kernel access; KCAPI_ACCESS_VMSPLICE - use vmsplice
+ *	    access; KCAPI_ACCESS_SENDMSG - sendmsg access)
  *
  * The AEAD cipher operation requires the furnishing of the associated
  * authentication data. In case such data is not required, it can be set to
@@ -1311,6 +1288,8 @@ void kcapi_aead_getdata(struct kcapi_handle *handle,
  * posix_memalign(PAGE_SIZE), If it is not aligned at the page boundary,
  * the vmsplice call may not send all data to the kernel.
  *
+ * The IV buffer must be exactly kcapi_cipher_ivsize() bytes in size.
+ *
  * To catch authentication errors (i.e. integrity violations) during the
  * decryption operation, the errno of this call shall be checked for EBADMSG.
  * If this function returns < 0 and errno is set to EBADMSG, an authentication
@@ -1321,8 +1300,9 @@ void kcapi_aead_getdata(struct kcapi_handle *handle,
  */
 ssize_t kcapi_aead_decrypt(struct kcapi_handle *handle,
 			   const unsigned char *in, size_t inlen,
+			   const unsigned char *iv,
 			   const unsigned char *assoc, const unsigned char *tag,
-			   unsigned char *out, size_t outlen)
+			   unsigned char *out, size_t outlen, int access)
 {
 	struct iovec iov[3];
 	ssize_t ret = 0;
@@ -1337,9 +1317,10 @@ ssize_t kcapi_aead_decrypt(struct kcapi_handle *handle,
 		return -EINVAL;
 	}
 
-#if 0
-	if (inlen + handle->aead.assoclen + handle->aead.taglen >= (1<<13s)) {
-#endif
+	handle->cipher.iv = iv;
+
+	if (access == KCAPI_ACCESS_HEURISTIC ||
+	    access == KCAPI_ACCESS_SENDMSG) {
 		if (assoc && handle->aead.assoclen) {
 			iov[0].iov_base = (void*)(uintptr_t)assoc;
 			iov[0].iov_len = handle->aead.assoclen;
@@ -1359,8 +1340,6 @@ ssize_t kcapi_aead_decrypt(struct kcapi_handle *handle,
 		}
 		if (0 > ret)
 			return ret;
-#if 0
-	/* using separate pages slows down splice below the speed of sendmsg */
 	} else {
 		size_t len = 0;
 
@@ -1393,16 +1372,14 @@ ssize_t kcapi_aead_decrypt(struct kcapi_handle *handle,
 		if (0 > ret)
 			return ret;
 	}
-#endif
 
-	iov[0].iov_base = (void*)(uintptr_t)out;
-	iov[0].iov_len = outlen;
-	return _kcapi_common_recv_data(handle, &iov[0], 1);
+	return _kcapi_common_read_data(handle, out, outlen);
 }
 
 /**
  * kcapi_aead_stream_init_enc() - start an encryption operation (stream)
  * @handle: cipher handle - input
+ * @iv: IV to be used for cipher operation - input
  * @iov: scatter/gather list with data to be encrypted. This is the pointer to
  *	 the first @iov entry if an array of @iov entries is supplied. See
  *	 sendmsg(2) for details on how @iov is to be used. This pointer may be
@@ -1430,12 +1407,16 @@ ssize_t kcapi_aead_decrypt(struct kcapi_handle *handle,
  * the complete associated data must be provided, followed by (ii) the
  * plaintext.
  *
+ * The IV buffer must be exactly kcapi_cipher_ivsize() bytes in size.
+ *
  * Return: number of bytes sent to the kernel upon success;
  *	   < 0 in case of error with errno set
  */
 ssize_t kcapi_aead_stream_init_enc(struct kcapi_handle *handle,
+				   const unsigned char *iv,
 				   struct iovec *iov, size_t iovlen)
 {
+	handle->cipher.iv = iv;
 	return _kcapi_common_send_meta(handle, iov, iovlen, ALG_OP_ENCRYPT,
 				       MSG_MORE);
 }
@@ -1443,6 +1424,7 @@ ssize_t kcapi_aead_stream_init_enc(struct kcapi_handle *handle,
 /**
  * kcapi_aead_stream_init_dec() - start a decryption operation (stream)
  * @handle: cipher handle - input
+ * @iv: IV to be used for cipher operation - input
  * @iov: scatter/gather list with data to be encrypted. This is the pointer to
  *	 the first @iov entry if an array of @iov entries is supplied. See
  *	 sendmsg(2) for details on how @iov is to be used. This pointer may be
@@ -1470,12 +1452,16 @@ ssize_t kcapi_aead_stream_init_enc(struct kcapi_handle *handle,
  * the complete associated data must be provided, followed by (ii) the
  * plaintext. For decryption, also (iii) the tag value must be sent.
  *
+ * The IV buffer must be exactly kcapi_cipher_ivsize() bytes in size.
+ *
  * Return: number of bytes sent to the kernel upon success;
  *	   < 0 in case of error with errno set
  */
 ssize_t kcapi_aead_stream_init_dec(struct kcapi_handle *handle,
+				   const unsigned char *iv,
 				   struct iovec *iov, size_t iovlen)
 {
+	handle->cipher.iv = iv;
 	return _kcapi_common_send_meta(handle, iov, iovlen, ALG_OP_DECRYPT,
 				       MSG_MORE);
 }
@@ -1564,11 +1550,13 @@ ssize_t kcapi_aead_stream_op(struct kcapi_handle *handle,
 			"AEAD operation: No buffer for output data provided\n");
 		return -EINVAL;
 	}
+#if 0
 	if (iovlen != 1) {
 		fprintf(stderr,
 			"AEAD operation: Output IOV must contain only one entry\n");
 		return -EINVAL;
 	}
+#endif
 	return _kcapi_common_recv_data(handle, iov, iovlen);
 }
 
