@@ -30,7 +30,13 @@ struct aead_sg_list {
 
 struct aead_ctx {
 	struct aead_sg_list tsgl;
-	struct af_alg_sgl rsgl;
+	/*
+	 * RSGL_MAX_ENTRIES is an artificial limit where user space at maximum
+	 * can cause the kernel to allocate RSGL_MAX_ENTRIES * ALG_MAX_PAGES
+	 * bytes
+	 */
+#define RSGL_MAX_ENTRIES ALG_MAX_PAGES
+	struct af_alg_sgl rsgl[RSGL_MAX_ENTRIES];
 
 	void *iv;
 
@@ -42,7 +48,6 @@ struct aead_ctx {
 	bool more;
 	bool merge;
 	bool enc;
-	bool trunc;
 
 	size_t aead_assoclen;
 	struct aead_request aead_req;
@@ -88,33 +93,6 @@ static void aead_put_sgl(struct sock *sk)
 	ctx->used = 0;
 	ctx->more = 0;
 	ctx->merge = 0;
-	ctx->trunc = 0;
-}
-
-static int aead_wait_for_wmem(struct sock *sk, unsigned flags)
-{
-	long timeout;
-	DEFINE_WAIT(wait);
-	int err = -ERESTARTSYS;
-
-	if (flags & MSG_DONTWAIT)
-		return -EAGAIN;
-
-	set_bit(SOCK_ASYNC_NOSPACE, &sk->sk_socket->flags);
-
-	for (;;) {
-		if (signal_pending(current))
-			break;
-		prepare_to_wait(sk_sleep(sk), &wait, TASK_INTERRUPTIBLE);
-		timeout = MAX_SCHEDULE_TIMEOUT;
-		if (sk_wait_event(sk, &timeout, aead_writable(sk))) {
-			err = 0;
-			break;
-		}
-	}
-	finish_wait(sk_sleep(sk), &wait);
-
-	return err;
 }
 
 static void aead_wmem_wakeup(struct sock *sk)
@@ -142,9 +120,8 @@ static int aead_wait_for_data(struct sock *sk, unsigned flags)
 	DEFINE_WAIT(wait);
 	int err = -ERESTARTSYS;
 
-	if (flags & MSG_DONTWAIT) {
+	if (flags & MSG_DONTWAIT)
 		return -EAGAIN;
-	}
 
 	set_bit(SOCK_ASYNC_WAITDATA, &sk->sk_socket->flags);
 
@@ -173,6 +150,8 @@ static void aead_data_wakeup(struct sock *sk)
 
 	if (ctx->more)
 		return;
+	if (!ctx->used)
+		return;
 
 	rcu_read_lock();
 	wq = rcu_dereference(sk->sk_wq);
@@ -185,7 +164,7 @@ static void aead_data_wakeup(struct sock *sk)
 }
 
 static int aead_sendmsg(struct kiocb *unused, struct socket *sock,
-		        struct msghdr *msg, size_t size)
+			struct msghdr *msg, size_t size)
 {
 	struct sock *sk = sock->sk;
 	struct alg_sock *ask = alg_sk(sk);
@@ -254,20 +233,14 @@ static int aead_sendmsg(struct kiocb *unused, struct socket *sock,
 			ctx->used += len;
 			copied += len;
 			size -= len;
+			continue;
 		}
 
 		if (!aead_writable(sk)) {
-			/*
-			 * If there is more data to be expected, but we cannot
-			 * write more data, forcefully define that we do not
-			 * expect more data to invoke the AEAD operation. This
-			 * prevents a deadlock in user space.
-			 */
-			ctx->more = 0;
-			ctx->trunc = 1;
-			err = aead_wait_for_wmem(sk, msg->msg_flags);
-			if (err)
-				goto unlock;
+			/* user space sent too much data */
+			aead_put_sgl(sk);
+			err = -EMSGSIZE;
+			goto unlock;
 		}
 
 		/* allocate a new page */
@@ -276,6 +249,7 @@ static int aead_sendmsg(struct kiocb *unused, struct socket *sock,
 			int plen = 0;
 
 			if (sgl->cur >= ALG_MAX_PAGES) {
+				aead_put_sgl(sk);
 				err = -E2BIG;
 				goto unlock;
 			}
@@ -345,12 +319,10 @@ static ssize_t aead_sendpage(struct socket *sock, struct page *page,
 		goto done;
 
 	if (!aead_writable(sk)) {
-		/* see aead_sendmsg why more is set to 0 */
-		ctx->more = 0;
-		ctx->trunc = 1;
-		err = aead_wait_for_wmem(sk, flags);
-		if (err)
-			goto unlock;
+		/* user space sent too much data */
+		aead_put_sgl(sk);
+		err = -EMSGSIZE;
+		goto unlock;
 	}
 
 	ctx->merge = 0;
@@ -385,36 +357,37 @@ static int aead_recvmsg(struct kiocb *unused, struct socket *sock,
 	unsigned bs = crypto_aead_blocksize(crypto_aead_reqtfm(&ctx->aead_req));
 	unsigned as = crypto_aead_authsize(crypto_aead_reqtfm(&ctx->aead_req));
 	struct aead_sg_list *sgl = &ctx->tsgl;
-	struct scatterlist *sg = sgl->sg;
+	struct scatterlist *sg = NULL;
 	struct scatterlist assoc[ALG_MAX_PAGES];
 	size_t assoclen = 0;
 	unsigned int i = 0;
 	int err = -EINVAL;
 	unsigned long used = 0;
 	unsigned long outlen = 0;
+	const struct iovec *iov;
+	unsigned long iovlen = 0;
+	unsigned long usedpages = 0;
 
-	/*
-	 * Require exactly one IOV block as the AEAD operation is a one shot
-	 * due to the authentication tag.
-	 */
-	if (msg->msg_iter.nr_segs != 1)
+	/* Limit number of IOV blocks to be accessed below */
+	if (msg->msg_iter.nr_segs > RSGL_MAX_ENTRIES)
 		return -ENOMSG;
 
 	lock_sock(sk);
+
 	/*
-	* AEAD memory structure: For encryption, the tag is appended to the
-	* ciphertext which implies that the memory allocated for the ciphertext
-	* must be increased by the tag length. For decryption, the tag
-	* is expected to be concatenated to the ciphertext. The plaintext
-	* therefore has a memory size of the ciphertext minus the tag length.
-	*
-	* The memory structure for cipher operation has the following
-	* structure:
-	*	AEAD encryption input:  assoc data || plaintext
-	*	AEAD encryption output: cipherntext || auth tag
-	*	AEAD decryption input:  assoc data || ciphertext || auth tag
-	*	AEAD decryption output: plaintext
-	*/
+	 * AEAD memory structure: For encryption, the tag is appended to the
+	 * ciphertext which implies that the memory allocated for the ciphertext
+	 * must be increased by the tag length. For decryption, the tag
+	 * is expected to be concatenated to the ciphertext. The plaintext
+	 * therefore has a memory size of the ciphertext minus the tag length.
+	 *
+	 * The memory structure for cipher operation has the following
+	 * structure:
+	 *	AEAD encryption input:  assoc data || plaintext
+	 *	AEAD encryption output: cipherntext || auth tag
+	 *	AEAD decryption input:  assoc data || ciphertext || auth tag
+	 *	AEAD decryption output: plaintext
+	 */
 
 	if (ctx->more) {
 		err = aead_wait_for_data(sk, flags);
@@ -454,17 +427,33 @@ static int aead_recvmsg(struct kiocb *unused, struct socket *sock,
 		outlen = ((outlen + bs - 1) / bs * bs);
 	}
 
-	/* ensure output buffer is sufficiently large */
-	if (msg->msg_iter.iov->iov_len < outlen)
-		goto unlock;
+	/* convert iovecs of output buffers into scatterlists */
+	for (iov = msg->msg_iter.iov, iovlen = 0;
+	     iovlen < msg->msg_iter.nr_segs; iovlen++, iov++) {
+		char __user *from = iov->iov_base;
+		unsigned long seglen = min_t(unsigned long, iov->iov_len,
+					     (outlen - usedpages));
 
-	outlen = af_alg_make_sg(&ctx->rsgl, msg->msg_iter.iov->iov_base,
-				outlen, 1);
-	err = outlen;
-	if (err < 0)
-		goto unlock;
+		/* make one iovec available as scatterlist */
+		err = af_alg_make_sg(&ctx->rsgl[iovlen], from, seglen, 1);
+		if (err < 0)
+			goto unlock;
+		usedpages += err;
+		/* chain the new scatterlist with initial list */
+		if (iovlen)
+			scatterwalk_crypto_chain(ctx->rsgl[0].sg,
+					ctx->rsgl[iovlen].sg, 1,
+					sg_nents(ctx->rsgl[iovlen-1].sg));
+		/* we do not need more iovecs as we have sufficient memory */
+		if (outlen <= usedpages)
+			break;
+	}
 
 	err = -EINVAL;
+	/* ensure output buffer is sufficiently large */
+	if (usedpages < outlen)
+		goto unlock;
+
 	sg_init_table(assoc, ALG_MAX_PAGES);
 	assoclen = ctx->aead_assoclen;
 	/*
@@ -504,19 +493,14 @@ static int aead_recvmsg(struct kiocb *unused, struct socket *sock,
 	}
 
 	aead_request_set_assoc(&ctx->aead_req, assoc, ctx->aead_assoclen);
-	aead_request_set_crypt(&ctx->aead_req, sg, ctx->rsgl.sg, used, ctx->iv);
+	aead_request_set_crypt(&ctx->aead_req, sg, ctx->rsgl[0].sg, used,
+			       ctx->iv);
 
 	err = af_alg_wait_for_completion(ctx->enc ?
 					 crypto_aead_encrypt(&ctx->aead_req) :
 					 crypto_aead_decrypt(&ctx->aead_req),
 					 &ctx->completion);
 
-	af_alg_free_sg(&ctx->rsgl);
-
-	/* indicate userspace that we processed incomplete data */
-	if (ctx->trunc)
-		msg->msg_flags |= MSG_TRUNC;
-
 	if (err) {
 		/* EBADMSG implies a valid cipher operation took place */
 		if (err == -EBADMSG)
@@ -529,6 +513,9 @@ static int aead_recvmsg(struct kiocb *unused, struct socket *sock,
 	err = 0;
 
 unlock:
+	for (i = 0; i < iovlen; i++)
+		af_alg_free_sg(&ctx->rsgl[i]);
+
 	aead_wmem_wakeup(sk);
 	release_sock(sk);
 
@@ -636,7 +623,6 @@ static int aead_accept_parent(void *private, struct sock *sk)
 	ctx->enc = 0;
 	ctx->tsgl.cur = 0;
 	ctx->aead_assoclen = 0;
-	ctx->trunc = 0;
 	af_alg_init_completion(&ctx->completion);
 	sg_init_table(ctx->tsgl.sg, ALG_MAX_PAGES);
 
